%{
(*
 * fc_ast_parse.mly -- Parser for functional C
 *
 * Ling Li, Xin Yu
 * Feb. 23, 2001, Caltech
 *)

open Fc_ast_type
open Fc_ast_util

(*
 * Build an assignment expression.
 *)
let make_asgnop op expr1 expr2 =
   let pos = union_pos (pos_of_expr expr1) (pos_of_expr expr2) in
      AssignExpr (op, expr1, expr2, pos)

let make_assign op expr1 expr2 =
   make_asgnop (Some op) expr1 expr2

(*
 * Build a binary expression.
 *)
let make_binop op expr1 expr2 =
   let pos = union_pos (pos_of_expr expr1) (pos_of_expr expr2) in
      BinOpExpr (op, expr1, expr2, pos)

(*
 * Build a boolean expression.
 *)
let make_boolop op expr1 expr2 =
   let pos = union_pos (pos_of_expr expr1) (pos_of_expr expr2) in
      BoolOpExpr (op, expr1, expr2, pos)

(*
 * Build a unary expression.
 *)
let make_unaryop op expr =
   let pos = pos_of_expr expr in
      UnOpExpr (op, expr, pos)

(*
 * Build an addr_of expression.
 *)
let make_addrofop expr =
   let pos = pos_of_expr expr in
      AddrOfExpr (expr, pos)

(*
 * Build a unary arithmetic expression.
 *)
let make_uarithop op expr =
   let pos = pos_of_expr expr in
      UArithExpr (op, expr, pos)

(*
 * Build a subscript expression.
 *)
let make_subscriptop expr1 expr2 =
   let pos = union_pos (pos_of_expr expr1) (pos_of_expr expr2) in
      SubscriptExpr (expr1, expr2, pos)

(*
 * Build a struct projection expression.
 *)
let make_structprojop expr id =
   let pos = union_pos (pos_of_expr expr) (snd id) in
      ProjectExpr (expr, fst id, pos)

(* Empty symbol, position, and type *)
let emsym = Symbol.new_symbol_string "n/a"
let empos = "n/a", -1, -1, -1, -1
let emtyp = TypeInt empos

(* Replace the 'deepest' empty type with type_pos *)
let rec fill_type typ_pos = function
 | TypeInt _ -> typ_pos
 | TypeArray (typ, exp, pos) -> TypeArray (fill_type typ_pos typ, exp, pos)
 | TypeFun (typ_l, typ, pos) -> TypeFun (typ_l, fill_type typ_pos typ, pos)
 | _ -> raise (Invalid_argument "Fc_ast_parse.fill_type: wrong type")

let map_type typ_pos decl_list =
   List.map (fun (s,t,e,p) -> (s,(fill_type typ_pos t),e,p)) decl_list

(* Add new-defined type into the type table *)
let add_typedefs (l : var_decl list) =
   List.iter (fun (v, _, _) -> Fc_ast_state.add_type v) l

(* Convert 'var_decl list' to 'ty list' *)
let abstract_type l = List.map (fun (_, t, _) -> t) l

(* 'union_pos' that can deal with empty position (empos) *)
let union_pos a b =
   if a = empos then b
   else if b = empos then a
   else union_pos a b
%}

/* Binary opertors */
%token <Fc_ast_type.pos> TokPlus                      /* + */
%token <Fc_ast_type.pos> TokMinus                     /* - */
%token <Fc_ast_type.pos> TokStar                      /* * */
%token <Fc_ast_type.pos> TokSlash                     /* / */
%token <Fc_ast_type.pos> TokMod                       /* % */

%token <Fc_ast_type.pos> TokIncr                      /* ++ */
%token <Fc_ast_type.pos> TokDecr                      /* -- */

%token <Fc_ast_type.pos> TokAssign                    /* = */
%token <Fc_ast_type.pos> TokPlusEq                    /* += */
%token <Fc_ast_type.pos> TokMinusEq                   /* -= */
%token <Fc_ast_type.pos> TokStarEq                    /* *= */
%token <Fc_ast_type.pos> TokSlashEq                   /* /= */
%token <Fc_ast_type.pos> TokModEq                     /* %= */
%token <Fc_ast_type.pos> TokSizeof                    /* sizeof */

/* Logical operators */
%token <Fc_ast_type.pos> TokLOr                       /* || */
%token <Fc_ast_type.pos> TokLAnd                      /* && */
%token <Fc_ast_type.pos> TokEq TokNotEq               /* ==, != */
%token <Fc_ast_type.pos> TokLe TokLt TokGe TokGt      /* <=, <, >=, > */
%token <Fc_ast_type.pos> TokNot                       /* ! */

/* Bitwise Operators */
%token <Fc_ast_type.pos> TokBOr  TokBOrEq             /* |, |= */
%token <Fc_ast_type.pos> TokBXor TokBXorEq            /* ^, ^= */
%token <Fc_ast_type.pos> TokBAnd TokBAndEq            /* &, &= */
%token <Fc_ast_type.pos> TokBNot                      /* ~ */
%token <Fc_ast_type.pos> TokLShift TokLShiftEq        /* <<, <<= */
%token <Fc_ast_type.pos> TokRShift TokRShiftEq        /* >>, >>= */

/* Keywords & built-in types */
%token <Fc_ast_type.pos> TokTypedef
%token <Fc_ast_type.pos> TokIf
%token <Fc_ast_type.pos> TokElse
%token <Fc_ast_type.pos> TokFor
%token <Fc_ast_type.pos> TokWhile
%token <Fc_ast_type.pos> TokReturn
%token <Fc_ast_type.pos> TokBreak
%token <Fc_ast_type.pos> TokStruct

%token <Fc_ast_type.pos> TokTypeInt
%token <Fc_ast_type.pos> TokTypeChar
%token <Fc_ast_type.pos> TokTypeBool
%token <Fc_ast_type.pos> TokTypeFloat

/* Symbols */
%token <Fc_ast_type.pos> TokSemi                      /* ; */
%token <Fc_ast_type.pos> TokComma                     /* , */
%token <Fc_ast_type.pos> TokQues                      /* ? */
%token <Fc_ast_type.pos> TokColon                     /* : */
%token <Fc_ast_type.pos> TokDot                       /* . */
%token <Fc_ast_type.pos> TokArrow                     /* -> */
%token <Fc_ast_type.pos> TokLeftParen                 /* ( */
%token <Fc_ast_type.pos> TokRightParen                /* ) */
%token <Fc_ast_type.pos> TokLeftBrack                 /* [ */
%token <Fc_ast_type.pos> TokRightBrack                /* ] */
%token <Fc_ast_type.pos> TokLeftBrace                 /* { */
%token <Fc_ast_type.pos> TokRightBrace                /* } */

/* Other stuff */
%token <Fc_ast_type.pos> TokEof

/* Names, numbers, etc. */
%token <Symbol.symbol * Fc_ast_type.pos> TokId
%token <Symbol.symbol * Fc_ast_type.pos> TokTypeId
%token <string * Fc_ast_type.pos> TokString
%token <char * Fc_ast_type.pos> TokChar
%token <int * Fc_ast_type.pos> TokInt
%token <float * Fc_ast_type.pos> TokFloat

/* Associate precedences and associativities */
%left  TokComma                                       /* , */
%right TokAssign TokLShiftEq TokRShiftEq              /* = <<= >>= */
       TokPlusEq TokMinusEq TokStarEq TokSlashEq      /* += -= *= /= */
       TokModEq TokBOrEq TokBAndEq TokBXorEq          /* %= |= &= ^= */
%right TokQues TokColon                               /* ? : */
%left  TokLOr                                         /* || */
%left  TokLAnd                                        /* && */
%left  TokBOr                                         /* | */
%left  TokBXor                                        /* ^ */
%left  TokBAnd                                        /* & */
%left  TokEq TokNotEq                                 /* == != */
%left  TokLe TokLt TokGe TokGt                        /* <= < >= > */
%left  TokLShift TokRShift                            /* << >> */
%left  TokPlus TokMinus                               /* + - */
%left  TokStar TokSlash TokMod                        /* * / % */
%right prec_unplus prec_unminus prec_ustar prec_uaddr /* + - * & */
       TokNot TokBNot TokIncr TokDecr TokSizeof       /* ! ~ ++ -- sizeof */
%left  TokLeftParen TokRightParen TokLeftBrack        /* () [] */
       TokRightBrack TokArrow TokDot                  /* -> . */

/*
 * Declare the "start" productions.
 * These declarations define functions for parsing
 * the productions.  You can define as many as you want.
 */
%start prog
%start topstmt
%type <Fc_ast_type.prog> prog
%type <Fc_ast_type.expr option> topstmt

%%

/*
 * This defines the valid contents of a file.
 */
prog:
 | stmt_list TokEof { $1 }
 ;

/*
 * Toplevel statement.  For evaluation,
 * we allow type, function, and value definitions
 * at the top level.
 */
topstmt:
 | stmt  {
      Format.print_string ">> ";
      print_expr $1;
      Format.print_string "\n";
      Some $1
   }
 | TokEof { None }
 ;

/*
 * Expressions.
 */
expr:
 | TokInt      { let i, pos = $1 in IntExpr (i, pos) }
 | TokChar     { let c, pos = $1 in CharExpr (c, pos) }
 | TokFloat    { let x, pos = $1 in FloatExpr (x, pos) }
 | string_list { let s, pos = $1 in StringExpr (s, pos) }

 | TokId       { let s, pos = $1 in VarExpr (s, pos) }

 | expr TokPlus   expr  { make_binop   PlusOp   $1 $3 }
 | expr TokMinus  expr  { make_binop   MinusOp  $1 $3 }
 | expr TokStar   expr  { make_binop   TimesOp  $1 $3 }
 | expr TokSlash  expr  { make_binop   DivideOp $1 $3 }
 | expr TokMod    expr  { make_binop   ModOp    $1 $3 }
 | expr TokBAnd   expr  { make_binop   BAndOp   $1 $3 }
 | expr TokBOr    expr  { make_binop   BOrOp    $1 $3 }
 | expr TokBXor   expr  { make_binop   BXorOp   $1 $3 }
 | expr TokLShift expr  { make_binop   LShiftOp $1 $3 }
 | expr TokRShift expr  { make_binop   RShiftOp $1 $3 }
 | expr TokEq     expr  { make_binop   EqOp     $1 $3 }
 | expr TokNotEq  expr  { make_binop   NotEqOp  $1 $3 }
 | expr TokLe     expr  { make_binop   LeOp     $1 $3 }
 | expr TokLt     expr  { make_binop   LtOp     $1 $3 }
 | expr TokGe     expr  { make_binop   GeOp     $1 $3 }
 | expr TokGt     expr  { make_binop   GtOp     $1 $3 }
 | expr TokLOr    expr  { make_boolop  LOrOp    $1 $3 }
 | expr TokLAnd   expr  { make_boolop  LAndOp   $1 $3 }
 | expr TokAssign expr  { make_asgnop  None     $1 $3 }
 | expr TokPlusEq expr  { make_assign  PlusOp   $1 $3 }
 | expr TokMinusEq expr { make_assign  MinusOp  $1 $3 }
 | expr TokStarEq expr  { make_assign  TimesOp  $1 $3 }
 | expr TokSlashEq expr { make_assign  DivideOp $1 $3 }
 | expr TokModEq  expr  { make_assign  ModOp    $1 $3 }
 | expr TokBAndEq expr  { make_assign  BAndOp   $1 $3 }
 | expr TokBOrEq  expr  { make_assign  BOrOp    $1 $3 }
 | expr TokBXorEq expr  { make_assign  BXorOp   $1 $3 }
 | expr TokLShiftEq expr{ make_assign  LShiftOp $1 $3 }
 | expr TokRShiftEq expr{ make_assign  RShiftOp $1 $3 }

 | TokIncr  expr  { make_uarithop PreIncrOp  $2 }
 | TokDecr  expr  { make_uarithop PreDecrOp  $2 }
 | expr  TokIncr  { make_uarithop PostIncrOp $1 }
 | expr  TokDecr  { make_uarithop PostDecrOp $1 }

 | TokMinus expr  { make_unaryop UMinusOp $2 }  %prec prec_unminus
 | TokPlus  expr  { $2 }                        %prec prec_unplus
 | TokNot   expr  { make_unaryop UNotOp   $2 }
/* | TokBNot  expr  { make unaryop UBNotOp  $2 } ???not implemented */
 | TokStar  expr  { make_unaryop UStarOp  $2 }  %prec prec_ustar
 | TokBAnd  expr  { make_addrofop         $2 }  %prec prec_uaddr

 | expr TokQues expr TokColon expr {
      let pos = union_pos (pos_of_expr $1) (pos_of_expr $5) in
         IfExpr ($1, $3, Some $5, pos)
   }
 | TokSizeof expr { raise (Invalid_argument "sizeof not implemented") }

/* ??? format below */
 | expr TokLeftBrack expr TokRightBrack   { make_subscriptop $1 $3 }
 | expr TokDot TokId    { make_structprojop $1 $3 }
 | expr TokArrow TokId  { make_structprojop (make_unaryop UStarOp $1) $3 }

 | TokLeftParen expr TokRightParen   { $2 }

 | expr TokLeftParen expr_list TokRightParen {
      let pos = union_pos (pos_of_expr $1) $4 in
         ApplyExpr ($1, $3, pos)
   }
 ;

/*
 * Merge string constants.
 * Example: "Hello " "World" -> "Hello World"
 */
string_list:
 | TokString   { $1 }
 | string_list TokString {
      let s1, pos1 = $1 in
      let s2, pos2 = $2 in
      (s1 ^ s2), (union_pos pos1 pos2)
   }
 ;

expr_list:
 | /* empty */             { [] }
 | expr                    { [$1] }
 | expr TokComma expr_list { $1 :: $3 }
 ;

/*
 * Some example statements.
 */
stmt:
 | def { $1 }
 | TokSemi
   { SeqExpr ([], $1) }
 | expr TokSemi
   { $1 }
 | TokIf TokLeftParen expr TokRightParen stmt TokElse stmt
   {  let pos = union_pos $1 (pos_of_expr $7) in
         IfExpr ($3, $5, Some $7, pos)
   }
 | TokIf TokLeftParen expr TokRightParen stmt
   {  let pos = union_pos $1 (pos_of_expr $5) in
         IfExpr ($3, $5, None, pos)
   }
 | TokWhile TokLeftParen expr TokRightParen stmt
   {  let pos = union_pos $1 (pos_of_expr $5) in
         WhileExpr ($3, $5, pos)
   }
 | TokFor TokLeftParen expr TokSemi expr TokSemi expr TokRightParen stmt
   {  let pos = union_pos $1 (pos_of_expr $9) in
         ForExpr ($3, $5, $7, $9, pos)
   }

 | TokReturn expr TokSemi
       { let pos = union_pos $1 $3 in
               ReturnExpr ($2, pos)
   }
 | TokBreak TokSemi {
      let pos = union_pos $1 $2 in
         BreakExpr pos
   }
 | TokLeftBrace stmt_list TokRightBrace {
      let pos = union_pos $1 $3 in
         SeqExpr ($2, pos)
   }
 ;

stmt_list:
 | /* empty */ {
      []
   }
 | stmt stmt_list {
      $1 :: $2
   }
 ;

/* Lots of other stuff here */

def:
 | var_defs    { $1 }
 | fun_def     { $1 }
 | TokTypedef type_defs_sub {
      let dl, pos = $2 in
      let pos = union_pos $1 pos in
         add_typedefs dl;
         TypeDefs (dl, pos)
   }
 | struct_def { $1 }
 ;

var_defs:
 | type_specifier init_declarator_list TokSemi {
      let pos = union_pos (pos_of_type $1) $3 in
         VarDefs (map_type $1 $2, pos)
   }
 ;

fun_def:
 | type_specifier declarator TokLeftBrace stmt_list TokRightBrace {
      let sym, typ, _, arg_list = $2 in
      let typ =
         match typ with
          | TypeFun (_, ty, _) -> fill_type $1 ty
          | _ -> raise (Invalid_argument "Fc_ast_parse.fun_def: wrong fun def")
      in
      let pos = union_pos (pos_of_type $1) $5 in
         FunDef (sym, arg_list, typ, SeqExpr ($4, (union_pos $3 $5)), pos)
   }
 ;

type_defs_sub:
 | type_specifier declarator_list TokSemi {
      let pos = union_pos (pos_of_type $1) $3 in
         (List.map (fun (s,t,p) -> (s,(fill_type $1 t),p)) $2), pos
   }
 ;

struct_def:
 | TokStruct TokId TokLeftBrace struct_declaration_list TokRightBrace
   init_declarator_list TokSemi {
      let sym, sym_pos = $2 in
      let pos = union_pos $1 $7 in
         Fc_ast_state.add_type sym;
         TypeDefs ([sym, TypeStruct ($4, pos), sym_pos], pos)
   }
 ;

type_specifier:
 | TokTypeChar    { TypeChar  $1 }
 | TokTypeInt     { TypeInt   $1 }
 | TokTypeFloat   { TypeFloat $1 }
 | TokTypeBool    { TypeBool  $1 }
 | struct_specifier { $1 }
 | TokTypeId {
      let sym, pos = $1 in
         TypeId (sym, pos)
   }
 ;

struct_specifier:
 | TokStruct TokLeftBrace struct_declaration_list TokRightBrace {
      let pos = union_pos $1 $4 in
         TypeStruct ($3, pos)
   }
 | TokStruct TokTypeId {
      (* 'struct typ var' is like 'typ var' *)
      (* check is needed: check whether t is a struct ???*)
      let sym, pos = $2 in
         TypeId (sym, pos)
   }

struct_declaration_list:
 | struct_declaration   { $1 }
 | struct_declaration struct_declaration_list {
      $1 @ $2
   }
 ;

struct_declaration:
 | type_defs_sub { fst $1 }
 ;


init_declarator_list:
 | /* empty */ {
      []
   }
 | init_declarator { [$1] }
 | init_declarator TokComma init_declarator_list {
      $1 :: $3
   }
 ;

init_declarator:     /* what is the pos of init_declarator? the pos of symbol or type? */
 | declarator {
      let sym, typ, sym_pos, _ = $1 in
         sym, typ, None, sym_pos
   }
 | declarator TokAssign expr {
      let sym, typ, sym_pos, _ = $1 in
         sym, typ, Some $3, sym_pos
   }
 ;

declarator_list:
 | declarator {
      let sym, typ, sym_pos, _ = $1 in
         [(sym, typ, sym_pos)]
   }
 | declarator TokComma declarator_list {
      let sym, typ, sym_pos, _ = $1 in
         (sym, typ, sym_pos) :: $3
   }

declarator:
 | direct_declarator { $1 }
 | TokStar declarator {
      let sym, typ, sym_pos, arg_list = $2 in
         let pos = union_pos $1 (pos_of_type typ) in
            sym, fill_type (TypeArray (emtyp, None, pos)) typ, sym_pos, arg_list
   }
 ;

direct_declarator:
 | TokId {
      let sym, pos = $1 in
         sym, emtyp, pos, []
   }
 | TokLeftParen declarator TokRightParen { $2 }
 | direct_declarator TokLeftBrack expr_list TokRightBrack {
      let sym, typ, sym_pos, arg_list = $1 in
      let pos = union_pos (pos_of_type typ) $4 in
      let len = List.length $3 in
      let dim =
         if len = 0 then None
         else Some (List.nth $3 (len-1))
      in
         sym, fill_type (TypeArray (emtyp, dim, pos)) typ, sym_pos, arg_list
   }
 | direct_declarator TokLeftParen parameter_list TokRightParen {
      let sym, typ, sym_pos, _ = $1 in
      let pos = union_pos (pos_of_type typ) $4 in
         sym, fill_type (TypeFun (abstract_type $3, emtyp, pos)) typ, sym_pos, $3
   }
 ;

parameter_list:
 | /* empty */ { [] }
 | parameter_declaration { [$1] }
 | parameter_declaration TokComma parameter_list { $1 :: $3 }
 ;

parameter_declaration:
 | type_specifier declarator {
      let sym, typ, pos, _ = $2 in
         sym, (fill_type $1 typ), pos
   }
 | type_specifier abstract_declarator {
      let sym, typ, pos, _ = $2 in
         sym, (fill_type $1 typ), pos
   }

abstract_declarator:
 | /* empty */ { emsym, emtyp, empos, [] }
 | abstract_declarator_noempty { $1 }
 ;

abstract_declarator_noempty:
 | direct_abstract_declarator { $1 }
 | TokStar {
      emsym, TypeArray (emtyp, None, $1), empos, []
   }
 | TokStar abstract_declarator_noempty {
      let sym, typ, sym_pos, arg_list = $2 in
      let pos = union_pos $1 (pos_of_type typ) in
         sym, fill_type (TypeArray (emtyp, None, pos)) typ, sym_pos, arg_list
   }
 ;

direct_abstract_declarator:
 | TokLeftParen abstract_declarator_noempty TokRightParen { $2 }
 | direct_abstract_declarator TokLeftBrack expr_list TokRightBrack {
      let sym, typ, sym_pos, arg_list = $1 in
      let pos = union_pos (pos_of_type typ) $4 in
      let len = List.length $3 in
      let dim =
         if len = 0 then None
         else Some (List.nth $3 (len-1))
      in
         sym, fill_type (TypeArray (emtyp, dim, pos)) typ, sym_pos, arg_list
   }
 | direct_abstract_declarator TokLeftParen parameter_list TokRightParen {
      let sym, typ, sym_pos, _ = $1 in
      let pos = union_pos (pos_of_type typ) $4 in
         sym, fill_type (TypeFun (abstract_type $3, emtyp, pos)) typ, sym_pos, $3
   }
 | TokLeftBrack expr_list TokRightBrack {
      let pos = union_pos $1 $3 in
      let len = List.length $2 in
      let dim =
         if len = 0 then None
         else Some (List.nth $2 (len-1))
      in
         emsym, TypeArray (emtyp, dim, pos), empos, []
   }
 | TokLeftParen parameter_list TokRightParen {
      let pos = union_pos $1 $3 in
         emsym, TypeFun (abstract_type $2, emtyp, pos), empos, $2
   }
 ;
