(*
 * Common-subexpression elimination.
 * This implementation is pretty simple:
 * subexpressions are the same only if they
 * compute exactly the same expression.
 *
 * Note: all variables are different.
 * IR must be functional.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn

(************************************************************************
 * ENVIRONMENTS
 ************************************************************************)

(*
 * You will need to define the type of variable environments.
 *)

(*
 * Add a abstract var to the venv.
 * This is a variable that can't be expanded to
 * an expression.
 *)
let venv_add_abs_var venv v =
   raise (Invalid_argument "venv_add_abs_var: NotImplemented")

(************************************************************************
 * CSE
 ************************************************************************)

(*
 * Common-subexpression elimination.
 * We use a simple algorithm: no memory optimization,
 * only identical code.
 *)
let rec cse_expr venv e =
   let pos = string_pos "cse_expr" (atom_pos e) in
      match e with
         LetAtom (v, ty, a, e) ->
            cse_atom_expr venv pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            cse_unop_expr venv pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            cse_binop_expr venv pos v ty op a1 a2 e
       | TailCall _
       | IntCall _ ->
            e
       | LetExtCall (v, f, ty, args, e) ->
            let venv = venv_add_abs_var venv v in
               LetExtCall (v, f, ty, args, cse_expr venv e)
       | IfThenElse (op, a1, a2, e1, e2) ->
            IfThenElse (op, a1, a2, cse_expr venv e1, cse_expr venv e2)
       | SetSubscript (a1, ty, a2, a3, e) ->
            SetSubscript (a1, ty, a2, a3, cse_expr venv e)
       | LetSubscript (v, ty, a1, a2, e) ->
            let venv = venv_add_abs_var venv v in
               LetSubscript (v, ty, a1, a2, cse_expr venv e)
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            cse_addr_of_subscript_expr venv pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            SetProject (a1, l, ty, a2, cse_expr venv e)
       | LetProject (v, ty, a, l, e) ->
            let venv = venv_add_abs_var venv v in
               LetProject (v, ty, a, l, cse_expr venv e)
       | LetAddrOfProject (v, ty, a, l, e) ->
            cse_addr_of_project_expr venv pos v ty a l e
       | Memcpy (a1, ty, a2, e) ->
            Memcpy (a1, ty, a2, cse_expr venv e)
       | LetAlloc (v, ty, e) ->
            let venv = venv_add_abs_var venv v in
               LetAlloc (v, ty, cse_expr venv e)
       | LetClosure (v, ty, f, args, e) ->
            cse_closure_expr venv pos v ty f args e
       | SetVar _
       | LetTypes _
       | LetFuns _
       | LetString _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            let pos = string_pos "cse_expr" (atom_pos e) in
               raise (IRException (pos, IRLevel 2))

(*
 * Inline an atom.
 *)
and cse_atom_expr venv pos v ty a e =
   let pos = string_pos "cse_atom_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Unary operation.
 *)
and cse_unop_expr venv pos v ty op a e =
   let pos = string_pos "cse_unop_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Binary operation.
 *)
and cse_binop_expr venv pos v ty op a1 a2 e =
   let pos = string_pos "cse_binop_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Address of subscript.
 *)
and cse_addr_of_subscript_expr venv pos v ty a1 a2 e =
   let pos = string_pos "cse_addr_of_subscript_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Address of struct projection.
 *)
and cse_addr_of_project_expr venv pos v ty a l e =
   let pos = string_pos "cse_addr_of_project_expr" pos in
      raise (IRException (pos, NotImplemented))

(*
 * Closure.
 *)
and cse_closure_expr venv pos v ty f args e =
   let pos = string_pos "cse_closure_expr" pos in
      raise (IRException (pos, NotImplemented))

(************************************************************************
 * GLOBAL FUNCTION
 ************************************************************************)

(*
 * Inline a prog.
 *)
let cse_prog prog =
   raise (Invalid_argument "cse_prog: NotImplemented")

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
