(*
 * Deadcode elimination.
 *
 * Ling Li, Xin Yu
 * Feb. 26, 2001, Caltech
 *)

open Debug
open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn

(************************************************************************
 * ENVIRONMENT
 ************************************************************************)

type env = SymbolSet.t

let env_empty =
   SymbolSet.empty

let env_add env v =
   SymbolSet.add v env

let env_union =
   SymbolSet.union

let env_equal =
   SymbolSet.equal

let env_mem env v =
   SymbolSet.mem v env

let env_remove env v =
   SymbolSet.remove v env

let env_remove_vars env vars =
   List.fold_left env_remove env vars

(************************************************************************
 * DEAD CODE ELIMINATION
 ************************************************************************)
let rec dead_type tenv = function
 | TyAny
 | TyChar
 | TyInt
 | TyFloat ->
      tenv
 | TyArray (ty, _) ->
      dead_type tenv ty
 | TyStruct fields ->
      List.fold_left (fun tenv (_, ty) -> dead_type tenv ty) tenv fields
 | TyFun (ty_vars, ty_res) ->
      let tenv = dead_type tenv ty_res in
         List.fold_left dead_type tenv ty_vars
 | TyId v ->
      env_add tenv v

let dead_atom tenv venv = function
 | AtomChar _
 | AtomInt _
 | AtomFloat _ ->
      tenv, venv
 | AtomNil ty ->
      dead_type tenv ty, venv
 | AtomVar v ->
      tenv, env_add venv v

let dead_atoms tenv venv atoms =
   List.fold_left (fun (tenv, venv) a ->
         dead_atom tenv venv a) (tenv, venv) atoms

(*
 * Remove dead expressions.
 *)
let rec dead_expr tenv venv e =
   let pos = string_pos "dead_expr" (atom_pos e) in
      match e with
         LetAtom (v, ty, a, e) ->
            dead_atom_expr tenv venv pos v ty a e
       | LetUnop (v, ty, op, a, e) ->
            dead_unop_expr tenv venv pos v ty op a e
       | LetBinop (v, ty, op, a1, a2, e) ->
            dead_binop_expr tenv venv pos v ty op a1 a2 e
       | TailCall (f, args) ->
            dead_tailcall_expr tenv venv pos f args
       | LetExtCall (v, f, ty, args, e) ->
            dead_extcall_expr tenv venv pos v f ty args e
       | IntCall (ty, f, args) ->
            dead_intcall_expr tenv venv pos ty f args
       | IfThenElse (op, a1, a2, e1, e2) ->
            dead_if_expr tenv venv pos op a1 a2 e1 e2
       | SetSubscript (a1, ty, a2, a3, e) ->
            dead_set_subscript_expr tenv venv pos ty a1 a2 a3 e
       | LetSubscript (v, ty, a1, a2, e) ->
            dead_subscript_expr tenv venv pos v ty a1 a2 e
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            dead_addr_of_subscript_expr tenv venv pos v ty a1 a2 e
       | SetProject (a1, l, ty, a2, e) ->
            dead_set_project_expr tenv venv pos ty a1 l a2 e
       | LetProject (v, ty, a, l, e) ->
            dead_project_expr tenv venv pos v ty a l e
       | LetAddrOfProject (v, ty, a, l, e) ->
            dead_addr_of_project_expr tenv venv pos v ty a l e
       | Memcpy (a1, ty, a2, e) ->
            dead_memcpy_expr tenv venv pos a1 ty a2 e
       | LetAlloc (v, ty, e) ->
            dead_alloc_expr tenv venv pos v ty e
       | LetClosure (v, ty, f, args, e) ->
            dead_closure_expr tenv venv pos v ty f args e
       | SetVar _
       | LetTypes _
       | LetFuns _
       | LetString _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            raise (IRException (pos, IRLevel 2))

(*
 * Elim this atom if the var is never used.
 *)
and dead_atom_expr tenv venv pos v ty a e =
   let pos = string_pos "dead_atom_expr" pos in
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a in
         let tenv = dead_type tenv ty in
            tenv, venv, LetAtom (v, ty, a, e)
      else
         tenv, venv, e

(*
 * Elim the unop if it is never used.
 *)
and dead_unop_expr tenv venv pos v ty op a e =
   (*let pos = string_pos "dead_unop_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a in
         let tenv = dead_type tenv ty in
            tenv, venv, LetUnop (v, ty, op, a, e)
      else
         tenv, venv, e

(*
 * Elim the binop if it is never used.
 *)
and dead_binop_expr tenv venv pos v ty op a1 a2 e =
   (*let pos = string_pos "dead_binop_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a1 in
         let tenv, venv = dead_atom tenv venv a2 in
         let tenv = dead_type tenv ty in
            tenv, venv, LetBinop (v, ty, op, a1, a2, e)
      else
         tenv, venv, e

(*
 * Base case.
 * Create the environments for the free vars.
 *)
and dead_tailcall_expr tenv venv pos f args =
   (*let pos = string_pos "dead_tailcall_expr" pos in*)
   let venv = env_add venv f in
   let tenv, venv = dead_atoms tenv venv args in
      tenv, venv, TailCall (f, args)

(*
 * External call.
 * Can't drop the function call, because it
 * may have a side-effect on the global vars.
 *)
and dead_extcall_expr tenv venv pos v f ty args e =
   (*let pos = string_pos "dead_extcall_expr" pos in*)
   let venv = env_add venv f in
   let tenv, venv = dead_atoms tenv venv args in
   let tenv = dead_type tenv ty in
   let tenv, venv, e = dead_expr tenv venv e in
   let venv = env_remove venv v in
      tenv, venv, LetExtCall (v, f, ty, args, e)

(*
 * Internal call.
 *)
and dead_intcall_expr tenv venv pos ty f args =
   (*let pos = string_pos "dead_intcall_expr" pos in*)
   let venv = env_add venv f in
   let tenv, venv = dead_atoms tenv venv args in
   let tenv = dead_type env_empty ty in
      env_empty, venv, IntCall (ty, f, args)

(*
 * Conditional is never dropped.
 *)
and dead_if_expr tenv venv pos op a1 a2 e1 e2 =
   (*let pos = string_pos "dead_if_expr" pos in*)
   let tenv, venv, e1 = dead_expr tenv venv e1 in
   let tenv, venv, e2 = dead_expr tenv venv e2 in
   let tenv, venv = dead_atom tenv venv a1 in
   let tenv, venv = dead_atom tenv venv a2 in
      tenv, venv, IfThenElse (op, a1, a2, e1, e2)

(*
 * Arrays.
 *)
and dead_subscript_expr tenv venv pos v ty a1 a2 e =
   (*let pos = string_pos "dead_subscript_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a1 in
         let tenv, venv = dead_atom tenv venv a2 in
         let tenv = dead_type tenv ty in
            tenv, venv, LetSubscript (v, ty, a1, a2, e)
      else
         tenv, venv, e

and dead_set_subscript_expr tenv venv pos ty a1 a2 a3 e =
   (*let pos = string_pos "dead_set_subscript_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
   let tenv, venv = dead_atom tenv venv a1 in
   let tenv, venv = dead_atom tenv venv a2 in
   let tenv, venv = dead_atom tenv venv a3 in
   let tenv = dead_type tenv ty in
      tenv, venv, SetSubscript (a1, ty, a2, a3, e)

and dead_addr_of_subscript_expr tenv venv pos v ty a1 a2 e =
   (*let pos = string_pos "dead_addr_of_subscript_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a1 in
         let tenv, venv = dead_atom tenv venv a2 in
         let tenv = dead_type tenv ty in
            tenv, venv, LetAddrOfSubscript (v, ty, a1, a2, e)
      else
         tenv, venv, e

(*
 * Structs.
 *)
and dead_project_expr tenv venv pos v ty a l e =
   (*let pos = string_pos "dead_project_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a in
         let tenv = dead_type tenv ty in
            tenv, venv, LetProject (v, ty, a, l, e)
      else
         tenv, venv, e

and dead_set_project_expr tenv venv pos ty a1 l a2 e =
   (*let pos = string_pos "dead_set_project_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
   let tenv, venv = dead_atom tenv venv a1 in
   let tenv, venv = dead_atom tenv venv a2 in
   let tenv = dead_type tenv ty in
      tenv, venv, SetProject (a1, l, ty, a2, e)

and dead_addr_of_project_expr tenv venv pos v ty a l e =
   (*let pos = string_pos "dead_addr_of_project_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv, venv = dead_atom tenv venv a in
         let tenv = dead_type tenv ty in
            tenv, venv, LetAddrOfProject (v, ty, a, l, e)
      else
         tenv, venv, e

(*
 * Memory operations.
 *)
and dead_memcpy_expr tenv venv pos a1 ty a2 e =
   (*let pos = string_pos "dead_memcpy_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
   let tenv, venv = dead_atom tenv venv a1 in
   let tenv, venv = dead_atom tenv venv a2 in
   let tenv = dead_type tenv ty in
      tenv, venv, Memcpy (a1, ty, a2, e)

and dead_alloc_expr tenv venv pos v ty e =
   (*let pos = string_pos "dead_alloc_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let tenv = dead_type tenv ty in
            tenv, venv, LetAlloc (v, ty, e)
      else
         tenv, venv, e

(*
 * Closure.
 *)
and dead_closure_expr tenv venv pos v ty f args e =
   (*let pos = string_pos "dead_closure_expr" pos in*)
   let tenv, venv, e = dead_expr tenv venv e in
      if env_mem venv v then
         let venv = env_remove venv v in
         let venv = env_add venv f in
         let tenv, venv = dead_atoms tenv venv args in
         let tenv = dead_type tenv ty in
            tenv, venv, LetClosure (v, ty, f, args, e)
      else
         tenv, venv, e

(************************************************************************
 * GLOBAL FUNCTIONS
 ************************************************************************)

(*
 * Program.
 *)
let dead_prog
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =
   let tenv = dead_type env_empty data_type in
   let venv = env_add env_empty init_sym in

   (* Remove dead codes in funtions *)
   let trans_fun (funs, f_env) (f, gflag, ty, vars, e) =
      let tenv, venv, e = dead_expr tenv venv e in
      let venv = env_remove_vars venv vars in
      let venv = env_remove venv f in
      let tenv = dead_type tenv ty in
      let funs = (f, gflag, ty, vars, e) :: funs in
      let f_env = (f, gflag, tenv, venv) :: f_env in
         funs, f_env
   in
   let funs, f_env = List.fold_left trans_fun ([], []) funs in

   (* Find all functions used by global functions *)
   let expand_env (tenv, venv) (f, gflag, tenv', venv') =
      if gflag || env_mem venv f then
         env_union tenv tenv', env_union venv venv'
      else
         tenv, venv
   in
   let rec find_live_env tenv venv =
      let tenv', venv' = List.fold_left expand_env (tenv, venv) f_env in
         if (env_equal tenv tenv') && (env_equal venv venv') then
            tenv, venv
         else
            find_live_env tenv' venv'
   in
   let tenv, venv = find_live_env tenv venv in

   (* Remove dead functions *)
   let is_fun_live (f, gflag, _, _, _) =
      gflag || env_mem venv f
   in
   let funs = List.rev (List.filter is_fun_live funs) in

   (* Remove dead types *)
   let types = List.filter (fun (v, _) -> env_mem tenv v) types in

   (* Remove dead strings *)
   let strings = List.filter (fun (v, _) -> env_mem venv v) strings in
      { prog_data_type = data_type;
        prog_init_sym = init_sym;
        prog_types = types;
        prog_strings = strings;
        prog_funs = funs
      }
(*??? All pos can be removed *)
