(*
 * FIR test function.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Debug
open Fc_ir_type
open Fc_ir_exn
open Fc_ir_eval

(*
 * Arguments.
 *)
let debug_print_ast = ref false
let debug_print_ir = ref false
let debug_check_ir = ref false
let debug_eval_ir = ref false
let stage = ref 100

let usage =
   Printf.sprintf "usage: compile [options] [files...] [-- <program arguments>]
-stage <num>: numbers are (default is %d):
\t0: AST only
\t1: IR conversion
\t2: CPS conversion
\t3: Optimization: LetAtom elimination
\t4: ESC conversion
\t5: Closure conversion
\t6: Convert to a program
\t7: Optimization: deadcode elimination
\t8: Optimization: function inlining
\t9: Optimization: common-subexpression elimination
\t10: Optimization: deadcode elimination (again)" !stage

(*
 * Program arguments.
 *)
let rev_argv = ref []

(*
 * Default name of the data segment.
 *)
let default_data_sym = Symbol.new_symbol_string "data_seg"
let default_data_type = TyStruct []

(*
 * The compiled program.
 *)
type program =
   NoneProg
 | ExprProg of var * ty * exp
 | ProgProg of prog

let compiled_prog = ref NoneProg

(*
 * Parser wrappers.
 *)
let parse_prog inx =
   Fc_ast_state.push_tenv ();
   let lexbuf = Lexing.from_channel inx in
   let prog = Fc_ast_parse.prog Fc_ast_lex.main lexbuf in
   let decls = Fc_ast_state.pop_tenv () in
      match decls with
         (_, _, pos) :: _ -> Fc_ast_type.TypeDefs (decls, pos) :: prog
       | [] -> prog

(*
 * Print and typecheck the IR.
 *)
let print_check_ir tenv venv stage exp =
   if !debug_print_ir then
      begin
         Format.set_formatter_out_channel stdout;
         Format.open_hvbox Fc_ast_util.tabstop;
         Format.print_string "Stage: ";
         Format.print_string stage;
         Format.print_space();
         Fc_ir_util.print_expr exp;
         Format.close_box ();
         Format.print_newline ()
      end;
   if !debug_check_ir then
      ignore (Fc_ir_typeof.check_expr tenv venv exp)

(*
 * Print and typecheck the IR.
 *)
let print_check_prog tenv venv stage prog =
   if !debug_print_ir then
      begin
         Format.set_formatter_out_channel stdout;
         Format.open_hvbox Fc_ast_util.tabstop;
         Format.print_string "Stage: ";
         Format.print_string stage;
         Format.print_space();
         Fc_ir_util.print_prog prog;
         Format.close_box ();
         Format.print_newline ()
      end;
   if !debug_check_ir then
      ignore (Fc_ir_typeof.check_prog tenv venv prog)

(*
 * Read a program.
 *)
let compile_prog prog =
   (* Print the program *)
   let _ =
      if !debug_print_ast then
         begin
            Format.set_formatter_out_channel stdout;
            Fc_ast_util.print_prog prog;
            Format.print_newline ()
         end
   in

   let level = !stage in
   let data_sym = default_data_sym in
   let data_type = default_data_type in
      (if level >= 1 then
          (* Produce the IR *)
          let tenv = Fc_ir_typeof.empty_tenv in
          let venv = Fc_ir_typeof.empty_venv in
          let _, _, ty, exp = Fc_ir_ast.make_prog tenv venv prog in
          let _ = print_check_ir tenv venv "Initial IR" exp in
             compiled_prog := ExprProg (data_sym, data_type, exp);

             (* Continuation-passing style *)
             if level >= 2 then
                let exp = Fc_ir_cont.cont_expr tenv exp in
                let _ = print_check_ir tenv venv "CPS conversion" exp in
                   compiled_prog := ExprProg (data_sym, data_type, exp);

                   (* Remove all LetAtom (just temporarily) *)
                   if level >= 3 then
                      let exp = Fc_ir_valias.inline_expr exp in
                      let _ = print_check_ir tenv venv "Remove variable aliases" in

                         (* Escape conversion *)
                         if level >= 4 then
                            let data_sym, data_type, exp = Fc_ir_esc.aelim_expr tenv exp in
                            let tenv = Fc_ir_env.tenv_add tenv data_sym data_type in
                            let venv = Fc_ir_env.venv_add venv data_sym (TyArray (TyId data_sym, None)) in
                            let _ = print_check_ir tenv venv "Escape analysis" exp in
                               compiled_prog := ExprProg (data_sym, data_type, exp);

                               (* Closure conversion *)
                               if level >= 5 then
                                  let exp = Fc_ir_closure.close_expr tenv venv data_sym data_type exp in
                                  let _ = print_check_ir tenv venv "Closure conversion" exp in
                                     compiled_prog := ExprProg (data_sym, data_type, exp);

                                     (* Convert to a program *)
                                     if level >= 6 then
                                        let prog = Fc_ir_prog.prog_of_expr data_sym data_type exp in
                                        let prog = Fc_ir_prog.prog_externals prog in
                                        let _ = print_check_prog tenv venv "Program closure" prog in
                                           compiled_prog := ProgProg prog;

                                           (* Dead-code elimination *)
                                           if level >= 7 then
                                              let prog = Fc_fir_dead.dead_prog prog in
                                              let _ = print_check_prog tenv venv "Deadcode elimination" prog in
                                                 compiled_prog := ProgProg prog;

                                                 (* Inlining *)
                                                 if level >= 8 then
                                                    let prog = Fc_fir_inline.inline_prog true prog in
                                                    let prog = Fc_fir_inline.inline_prog true prog in
                                                    let prog = Fc_fir_inline.inline_prog false prog in
                                                    let _ = print_check_prog tenv venv "Inlining" prog in
                                                       compiled_prog := ProgProg prog;

                                                       (* Common-subexpression elimination *)
                                                       if level >= 9 then
                                                          let prog = Fc_fir_cse.cse_prog prog in
                                                          let prog = Fc_fir_inline.inline_prog false prog in

                                                             (* Another round of dead-code elimination *)
                                                             if level >= 10 then
                                                                let prog = Fc_fir_dead.dead_prog prog in
                                                                let _ = print_check_prog tenv venv "Deadcode elimination" prog in
                                                                   compiled_prog := ProgProg prog)

(*
 * Compilation.
 *)
let compile name =
   (* Open the input channel and set the position *)
   let inx = open_in name in
   let _ = Fc_ast_state.set_current_position (name, 1, 0, 1, 0) in
      try
         let prog = parse_prog inx in
            compile_prog prog;
            close_in inx
      with
         exn ->
            Fc_ir_exn.print_exn_chan stderr exn;
            close_in inx;
            exit 1

(*
 * Evaluate the main function.
 *)
let call_main () =
   if !debug_eval_ir then
      let argv = Array.of_list (List.rev !rev_argv) in
      let data_sym, exp =
         match !compiled_prog with
            NoneProg ->
               default_data_sym, Return (AtomInt 0)
          | ExprProg (data_sym, data_type, e) ->
               let e =
                  LetTypes ([data_sym, data_type],
                            LetAlloc (data_sym, data_type, e))
               in
                  data_sym, e
          | ProgProg prog ->
               let data_sym, e = Fc_ir_prog.expr_of_prog prog in
                  data_sym, e
      in
      let tenv = Fc_ir_typeof.empty_tenv in
      let venv = Fc_ir_eval.empty_venv in
      let tenv, venv, v = Fc_ir_eval.eval_expr tenv venv data_sym exp in
         if Array.length argv = 0 then
            begin
               eprintf "You have to specify arguments with the -- option%t" eflush;
               exit 1
            end;
         ignore (Fc_ir_eval.eval_main tenv venv data_sym argv)

(*
 * Call the main program.
 *)
let spec =
   ["-print_ast", Arg.Unit (fun () -> debug_print_ast := true), "print expressions";
    "-print_ir", Arg.Unit (fun () -> debug_print_ir := true), "print intermediate code";
    "-check_ir", Arg.Unit (fun () -> debug_check_ir := true), "typecheck intermediate code";
    "-eval", Arg.Unit (fun () -> debug_eval_ir := true), "evaluate intermediate code";
    "-debug_eval", Arg.Unit (fun () ->
          debug_eval_ir := true;
          debug_eval := true), "use the debugger during evaluation";
    "-debug_pos", Arg.Unit (fun () -> debug_pos := true), "print position information";
    "-debug_closure", Arg.Unit (fun () -> Fc_ir_closure.debug_closure := true), "print closure equations";
    "-stage", Arg.Int (fun i -> stage := i), "set compiler stages";
    "--", Arg.Rest (fun s -> rev_argv := s :: !rev_argv), "program arguments"]

let _ =
   Arg.parse spec compile usage;
   Fc_ir_exn.catch call_main ()

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
