(*
 * Convert an expression to a program.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type
open Fc_ir_exn_type
open Fc_ir_exn
open Fc_ir_env
open Fc_ir_standardize

(*
 * Collect the info about the main program.
 *)
type info =
   { info_types : (var * ty) list;
     info_funs : fundef list;
     info_strings : (var * string) list
   }

(*
 * Default exception.
 *)
let raise_exn pos v =
   IRException (pos, v)

let tenv_lookup = tenv_lookup raise_exn
let tenv_expand = tenv_expand raise_exn

(*
 * Expand a function type.
 *)
let dest_opt_fun_type tenv pos ty =
   match tenv_expand tenv pos ty with
      TyFun (ty_vars, ty_res) -> Some (ty_vars, ty_res)
    | _ -> None

let dest_fun_type tenv pos ty =
   match dest_opt_fun_type tenv pos ty with
      Some (ty_vars, ty_res) -> ty_vars, ty_res
    | None -> raise (IRException (pos, NotAFunction ty))

(*
 * Exit program.
 *)
let exit_sym = Symbol.add "exit"

(*
 * Flatten the expression.
 *)
let rec collect_info info e =
   let pos = string_pos "collect_types" (atom_pos e) in
      match e with
         LetTypes (types', e) ->
            let info = { info with info_types = types' @ info.info_types } in
               collect_info info e
       | LetFuns (funs, e) ->
            let info, funs = collect_info_funs info funs in
            let info = { info with info_funs = funs @ info.info_funs } in
               collect_info info e
       | LetString (v, s, e) ->
            let info = { info with info_strings = (v, s) :: info.info_strings } in
               collect_info info e
       | LetAtom (v, ty, a, e) ->
            let info, e = collect_info info e in
               info, LetAtom (v, ty, a, e)
       | LetUnop (v, ty, op, a, e) ->
            let info, e = collect_info info e in
               info, LetUnop (v, ty, op, a, e)
       | LetBinop (v, ty, op, a1, a2, e) ->
            let info, e = collect_info info e in
               info, LetBinop (v, ty, op, a1, a2, e)
       | TailCall _
       | IntCall _ ->
            info, e
       | LetExtCall (v, f, ty, args, e) ->
            let info, e = collect_info info e in
               info, LetExtCall (v, f, ty, args, e)
       | IfThenElse (op, a1, a2, e1, e2) ->
            let info, e1 = collect_info info e1 in
            let info, e2 = collect_info info e2 in
               info, IfThenElse (op, a1, a2, e1, e2)
       | SetSubscript (a1, ty, a2, a3, e) ->
            let info, e = collect_info info e in
               info, SetSubscript (a1, ty, a2, a3, e)
       | LetSubscript (v, ty, a1, a2, e) ->
            let info, e = collect_info info e in
               info, LetSubscript (v, ty, a1, a2, e)
       | LetAddrOfSubscript (v, ty, a1, a2, e) ->
            let info, e = collect_info info e in
               info, LetAddrOfSubscript (v, ty, a1, a2, e)
       | SetProject (a1, l, ty, a2, e) ->
            let info, e = collect_info info e in
               info, SetProject (a1, l, ty, a2, e)
       | LetProject (v, ty, a, l, e) ->
            let info, e = collect_info info e in
               info, LetProject (v, ty, a, l, e)
       | LetAddrOfProject (v, ty, a, l, e) ->
            let info, e = collect_info info e in
               info, LetAddrOfProject (v, ty, a, l, e)
       | Memcpy (a1, ty, a2, e) ->
            let info, e = collect_info info e in
               info, Memcpy (a1, ty, a2, e)
       | LetAlloc (v, ty, e) ->
            let info, e = collect_info info e in
               info, LetAlloc (v, ty, e)
       | LetClosure (v, ty, f, args, e) ->
            let info, e = collect_info info e in
               info, LetClosure (v, ty, f, args, e)
       | SetVar _
       | LetAddrOfVar _
       | LetCopy _
       | LetApply _
       | Return _ ->
            raise (IRException (pos, IRLevel 1))

(*
 * Collect the function definitions.
 *)
and collect_info_funs info = function
   (f, gflag, ty, vars, e) :: funs ->
      let info, funs = collect_info_funs info funs in
      let info, e = collect_info info e in
      let fund = f, gflag, ty, vars, e in
         info, fund :: funs
 | [] ->
      info, []

(*
 * Produce the program.
 *)
let prog_of_expr data_sym data_type e =
   let pos = string_pos "make_prog" (atom_pos e) in
   let info =
      { info_types = [];
        info_funs = [];
        info_strings = []
      }
   in
   let info, e = collect_info info e in
   let { info_types = types;
         info_funs = funs;
         info_strings = strings
       } = info
   in
   let types = (data_sym, data_type) :: types in

   (* Initializer program *)
   let init_sym = new_symbol_string "init" in
   let init_type = TyFun ([TyArray (data_type, None); TyFun ([TyInt], TyInt)], TyInt) in
   let fund = init_sym, true, init_type, [data_sym; exit_sym], e in
   let funs = fund :: funs in
      (* Build the prog *)
      { prog_data_type = data_type;
        prog_init_sym = init_sym;
        prog_types = types;
        prog_strings = strings;
        prog_funs = funs
      }

(*
 * Add a wrapper for a global fun.
 * We have to figure out the return type from the
 * continuation argument.  The very first argument that
 * is a function is the continuation.
 *)
let add_global_fun tenv f ty vars e =
   (* Add global wrapper as an internal function call *)
   let pos = atom_pos (LetFuns ([f, true, ty, vars, e], Return (AtomInt 0))) in
   let pos = string_pos "add_global_fun" pos in
   let ty_vars, _ = dest_fun_type tenv pos ty in
      match ty_vars, vars with
         _ :: ty_cont :: ty_vars, _ :: _ :: vars ->
            (match dest_fun_type tenv pos ty_cont with
                [ty_res], TyInt ->
                   let v = new_symbol_string "int_call_arg" in
                   let ty = TyFun (ty_vars, ty_res) in
                   let args = List.map (fun v -> AtomVar v) vars in
                   let f' = Symbol.add ("%" ^ Symbol.to_string f) in
                      f', true, ty, vars, IntCall (ty_res, f, args)
              | _ ->
                   raise (IRException (pos, InternalError "illegal global function definition")))
       | _ ->
            raise (IRException (pos, InternalError "illegal global function definition"))

(*
 * Add wrappers for all global funs.
 *)
let rec add_global_funs tenv funs' = function
   fund :: funs ->
      let f, gflag, ty, vars, e = fund in
         if gflag then
            let fun1 = add_global_fun tenv f ty vars e in
            let fund = f, false, ty, vars, e in
               add_global_funs tenv (fun1 :: fund :: funs') funs
         else
            add_global_funs tenv (fund :: funs') funs
 | [] ->
      funs'

(*
 * Add external definitions.
 *)
let prog_externals prog =
   let tenv =
      List.fold_left (fun tenv (v, ty) ->
            tenv_add tenv v ty) empty_tenv prog.prog_types
   in
   let funs = add_global_funs tenv [] prog.prog_funs in
   let prog = { prog with prog_funs = funs } in
      standardize_prog prog

(*
 * Produce an expression from the program.
 *)
let expr_of_prog
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =
   let v = new_symbol_string "data_seg" in
   let e = TailCall (init_sym, [AtomVar v; AtomVar exit_sym]) in
   let e = LetFuns (funs, e) in
   let e =
      List.fold_left (fun e (v, s) ->
            LetString (v, s, e)) e strings
   in
   let e = LetAlloc (v, data_type, e) in
      v, LetTypes (types, e)

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
